#include <stdint.h>
#include <string.h>
#include <vector>

#include <stdexcept>

#include "s10_app.h"
#include "s10_ble.h"
#include "ble_srv_common.h"

#include "nordic_common.h"
#include "nrf.h"
#include "sdk_common.h"
#include "app_error.h"

#include "nrf_log.h"
#include "ble.h"
#include "ble_gap.h"
#include "ble_err.h"
#include "ble_hci.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_qwr.h"

#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "app_button.h"
#include "app_timer.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"


#define DEVICE_NAME "XXXXX" /**< Name of device. Will be included in the advertising data. */

#define APP_BLE_OBSERVER_PRIO 3 /**< Application's BLE observer priority. You shouldn't need to modify this value. */
#define APP_BLE_CONN_CFG_TAG 1  /**< A tag identifying the SoftDevice BLE configuration. */

#define APP_ADV_INTERVAL MSEC_TO_UNITS(1000, UNIT_0_625_MS) /**< The advertising interval (in units of 0.625 ms;). */
#define APP_ADV_DURATION BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED /**< The advertising time-out (in units of seconds). When set to 0, we will never time out. */

/* 
https://infocenter.nordicsemi.com/index.jsp?topic=%2Fdrivers_nrfx_v2.3.0%2Fgroup__nrfx__saadc.html
If both conn_sup_timeout and max_conn_interval are specified, 
then the following constraint applies: conn_sup_timeout * 4 > (1 + slave_latency) * max_conn_interval 
that corresponds to the following Bluetooth Spec requirement: 
The Supervision_Timeout in milliseconds shall be larger than 
    (1 + Conn_Latency) * Conn_Interval_Max * 2, 
    where Conn_Interval_Max is given in milliseconds.

// General Connection Parameter Request Guidelines:
//
// 1. Peripheral Latency:
//    - Should be less than or equal to 30 connection intervals.
//
// 2. Supervision Timeout:
//    - Should be between 6 seconds and 18 seconds.
//
// 3. Connection Interval:
//    - Interval Min should be at least 15 ms.
//    - Interval Min should not exceed 2 seconds.
//    - Interval Min should be a multiple of 15 ms.
//
// 4. Interval Max:
//    - Should be at least 15 ms greater than Interval Min, or
//    - Both Interval Max and Interval Min should be 15 ms.
//
// 5. Interval Max * (Peripheral Latency + 1):
//    - Should be 6 seconds or less.
//
// 6. Supervision Timeout:
//    - Should be greater than Interval Max * (Peripheral Latency + 1) * 3.
//
// Examples:
// - If Interval Min is set to 30 ms, Interval Max should be at least 45 ms.
// - If Interval Min is set to 15 ms, Interval Max can also be 15 ms.
// - If Peripheral Latency is 10 and Interval Max is 100 ms, then:
//   - Interval Max * (Peripheral Latency + 1) = 100 ms * 11 = 1100 ms (1.1 seconds), which is less than 6 seconds.
//   - Supervision Timeout should be greater than 3.3 seconds (1100 ms * 3).
*/

#define MIN_CONN_INTERVAL MSEC_TO_UNITS(150, UNIT_1_25_MS)
#define MAX_CONN_INTERVAL MSEC_TO_UNITS(450, UNIT_1_25_MS) 

#define SLAVE_LATENCY 0                                    /**< Slave latency. */
#define CONN_SUP_TIMEOUT MSEC_TO_UNITS(17000, UNIT_10_MS)   /**< Connection supervisory time-out (4 seconds). */


#define FIRST_CONN_PARAMS_UPDATE_DELAY APP_TIMER_TICKS(2000) /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (15 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY APP_TIMER_TICKS(5000)   /**< Time between each call to sd_ble_gap_conn_param_update after the first call (5 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT 50                        /**< Number of attempts before giving up the connection parameter negotiation. */

#define DEAD_BEEF 0xDEADBEEF /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID; /**< Handle of the current connection. */
uint8_t m_adv_handle = BLE_GAP_ADV_SET_HANDLE_NOT_SET;           /**< Advertising handle used to identify an advertising set. */
uint8_t m_enc_advdata[BLE_GAP_ADV_SET_DATA_SIZE_MAX];            /**< Buffer for storing an encoded advertising set. */
uint8_t m_enc_scan_response_data[BLE_GAP_ADV_SET_DATA_SIZE_MAX]; /**< Buffer for storing an encoded scan data. */

NRF_BLE_GATT_DEF(m_gatt); /**< GATT module instance. */
NRF_BLE_QWR_DEF(m_qwr);   /**< Context for the Queued Write module. */

void *m_ble_obs_ptr;
NRF_SDH_BLE_OBSERVER(m_ble_obs, BLE_NUS_BLE_OBSERVER_PRIO, s10::BLE::on_ble_service_event, &m_ble_obs_ptr);

static uint32_t packages_sent_counter = 0;

using namespace s10;

ble_gatts_char_handles_t Services::char_handles = {0};

void Services::send_and_delete_queue()
{
    while(!msg_queue.empty())
    {
        // fetch next and pop it
        Msg *msg = msg_queue.back();
        msg_queue.pop_back();

        uint16_t length = msg->size();
        ble_gatts_hvx_params_t params = {0};
        params.handle = char_handles.value_handle;
        params.type = BLE_GATT_HVX_NOTIFICATION;
        params.offset = 0;
        params.p_len = &length;
        params.p_data = msg->data();

        uint32_t err_code = sd_ble_gatts_hvx(m_conn_handle, &params);
        if (err_code == NRF_SUCCESS)
        {
            packages_sent_counter++; // Increment the sent counter
            NRF_LOG_INFO("BLE message sent: len=%d", length);
        }
        else
        {
            NRF_LOG_INFO("BLE message failed: %s", nrf_strerror_find(err_code));
        }
        delete msg;
    }

    Services::msg_queue.clear();
}


void ConnParams::error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}

ConnParams::ConnParams()
{
    memset(&this->cp_init, 0, sizeof(cp_init));

    this->cp_init.p_conn_params = NULL;
    this->cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    this->cp_init.next_conn_params_update_delay = NEXT_CONN_PARAMS_UPDATE_DELAY;
    this->cp_init.max_conn_params_update_count = MAX_CONN_PARAMS_UPDATE_COUNT;
    this->cp_init.start_on_notify_cccd_handle = BLE_GATT_HANDLE_INVALID;
    this->cp_init.disconnect_on_fail = false; // TOOD: should we disconnect on fail? otherwise we may be in a situation where the phone deos not support the bandwidth which may lead to ux problem.
    this->cp_init.evt_handler = ConnParams::on_event;
    this->cp_init.error_handler = ConnParams::error_handler;
}

void ConnParams::init(void)
{
    ret_code_t err_code;
    err_code = ble_conn_params_init(&this->cp_init);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for starting advertising.
 */
void Advertising::start(void)
{
    ret_code_t err_code;

    err_code = sd_ble_gap_adv_start(m_adv_handle, APP_BLE_CONN_CFG_TAG);
    APP_ERROR_CHECK(err_code);

    // automatically inherited to the connection tx
    err_code = sd_ble_gap_tx_power_set(BLE_GAP_TX_POWER_ROLE_ADV, m_adv_handle, RADIO_TXPOWER_TXPOWER_Pos8dBm);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_INFO("Advertising start.");
}


BLE::BLE() :
   services(Services::instance()), adv(Services::instance())
{


}

void BLE::on_ble_service_event(ble_evt_t const *p_ble_evt, void *p_context)
{
    auto _this = BLE::instance();

    if(_this) {
       switch(p_ble_evt->header.evt_id)
        {
            case BLE_GATTS_EVT_WRITE:
            {
                ble_gatts_evt_write_t const * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

                if ((p_evt_write->handle == _this->services->char_handles.cccd_handle) && (p_evt_write->len == BLE_CCCD_VALUE_LEN)) {
                        if (ble_srv_is_notification_enabled(p_evt_write->data)) {
                            NRF_LOG_INFO("Notification enabled");
                            _this->services->client_notifications_enabled = true;
                            App::instance()->dispatch(App::Event(App::Event::Type::CONNECTED)); // we are now really connected and ready to send

                        } else {
                            NRF_LOG_INFO("Notification disabled");
                        }

                } else {

                    // Access the written data
                    // uint16_t len = p_evt_write->len;
                    uint8_t const *data = p_evt_write->data;
                    if (data)
                    {
                        Services::MsgType type = (Services::MsgType)data[0];

                        switch (type)
                        {
                        case Services::MsgType::XXXXXXXXXX:
                        {
                            NRF_LOG_INFO("XXXXXXXXXXXXX");
                            App::instance()->dispatch(App::Event(App::Event::Type::XXXXXXXXXXx));
                            break;
                        }                        }
                    }
                }
                break;

            } // BLE_GATTS_EVT_WRITE
            default: {
                // Handle other BLE events here
                break;
            }
        }
    }

    return;

}
bool s10::BLE::is_connected()
{
    if(m_conn_handle != BLE_CONN_HANDLE_INVALID && services->client_notifications_enabled) {
        return true;
    }
    return false;
}
void BLE::on_ble_event(ble_evt_t const *p_ble_evt, void *p_context)
{
    auto err_code = NRF_SUCCESS;

    auto _this = *((BLE**)p_context);

    switch (p_ble_evt->header.evt_id)
    {
    case BLE_GAP_EVT_CONNECTED:
        NRF_LOG_INFO("BLE_GAP_EVT_CONNECTED");
        //led_on();
        m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
        _this->services->client_notifications_enabled = false; // initialize connection
        _this->state_ = BLE::State::CONNECTED;

        err_code = nrf_ble_qwr_conn_handle_assign(&m_qwr, m_conn_handle);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GAP_EVT_DISCONNECTED:
        NRF_LOG_INFO("BLE_GAP_EVT_DISCONNECTED");
        //led_off();
        m_conn_handle = BLE_CONN_HANDLE_INVALID;
        APP_ERROR_CHECK(err_code);
        _this->adv.start();
        _this->services->client_notifications_enabled = false;
        _this->state_ = BLE::State::ADVERTISING;
        App::instance()->dispatch(App::Event::Type::DISCONNECTED); // we are now really connected and ready to send

        break;

    case BLE_GATTS_EVT_HVN_TX_COMPLETE:
    {
        packages_sent_counter -= p_ble_evt->evt.gatts_evt.params.hvn_tx_complete.count;
        NRF_LOG_INFO("BLE_GATTS_EVT_HVN_TX_COMPLETE: unconfirmed: %d", packages_sent_counter);
    }
    break;

    case BLE_GAP_EVT_CONN_PARAM_UPDATE:
    {
        NRF_LOG_INFO("BLE_GAP_EVT_CONN_PARAM_UPDATE");
        const ble_gap_evt_t *gap_evt = &p_ble_evt->evt.gap_evt;
        NRF_LOG_INFO("Connection interval updated: Slave Latency: %d, Sup timeout: %d ms, Min: %d ms, Max: %d ms.",
            gap_evt->params.conn_param_update.conn_params.slave_latency,
            gap_evt->params.conn_param_update.conn_params.conn_sup_timeout*10,
            gap_evt->params.conn_param_update.conn_params.min_conn_interval*1.25,
            gap_evt->params.conn_param_update.conn_params.max_conn_interval*1.25);

    } break;


    case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
    {
        NRF_LOG_INFO("BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST");

        const ble_gap_evt_t *gap_evt = &p_ble_evt->evt.gap_evt;
        err_code = sd_ble_gap_conn_param_update(gap_evt->conn_handle,
                                                &gap_evt->params.conn_param_update_request.conn_params);

        if(err_code != NRF_SUCCESS) {
            NRF_LOG_ERROR("sd_ble_gap_conn_param_update failed: %d", err_code);
        }

        NRF_LOG_INFO("Connection interval update request: Slave Latency: %d, Sup timeout: %d ms, Min: %d ms, Max: %d ms.",
            gap_evt->params.conn_param_update.conn_params.slave_latency,
            gap_evt->params.conn_param_update.conn_params.conn_sup_timeout*10,
            gap_evt->params.conn_param_update.conn_params.min_conn_interval*1.25,
            gap_evt->params.conn_param_update.conn_params.max_conn_interval*1.25);
    }
    break;

    case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
        NRF_LOG_INFO("Received BLE_GAP_EVT_CONN_PARAM_REQUEST");
        // Pairing not supported
        err_code = sd_ble_gap_sec_params_reply(m_conn_handle,
                                               BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP,
                                               NULL,
                                               NULL);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
    {
        NRF_LOG_INFO("BLE_GAP_EVT_PHY_UPDATE_REQUEST");
        ble_gap_phys_t const phys =
            {
                .tx_phys = BLE_GAP_PHY_AUTO,
                .rx_phys = BLE_GAP_PHY_AUTO,
            };
        err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
        APP_ERROR_CHECK(err_code);
    }
    break;

    case BLE_GATTS_EVT_SYS_ATTR_MISSING:
        // No system attributes have been stored.
        err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0, 0);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GATTC_EVT_TIMEOUT:
        // Disconnect on GATT Client timeout event.
        NRF_LOG_DEBUG("GATT Client Timeout.");
        err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                         BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GATTS_EVT_TIMEOUT:
        // Disconnect on GATT Server timeout event.
        NRF_LOG_DEBUG("GATT Server Timeout.");
        err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                         BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
        APP_ERROR_CHECK(err_code);
        break;

    default:
        // No implementation needed.
        break;
    }
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
void BLE::init(void)
{
    ret_code_t err_code;

    m_ble_obs_ptr = this;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    static auto _this = this;
    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, BLE::on_ble_event, &_this);

    // initialize the rest of the BLE stack
    gap.init();
    gatt.init();
    services->init();
    adv.init();
    params.init();



    NRF_LOG_INFO("BLE::init() done");
}


void BLE::start()
{
    adv.start();
    state_ = State::ADVERTISING;
}


void ConnParams::on_event(ble_conn_params_evt_t *p_evt)
{
    ret_code_t err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}

QWR::QWR()
{
}

void QWR::init() {

    ret_code_t err_code;
    //s10_ble_init_t init = {0};
    nrf_ble_qwr_init_t qwr_init = {0};

    // Initialize Queued Write Module.
    qwr_init.error_handler = nrf_qwr_error_handler;

    err_code = nrf_ble_qwr_init(&m_qwr, &qwr_init);
    APP_ERROR_CHECK(err_code);

}

/**@brief Function for initializing services that will be used by the application.
 */
Services::Services()
{
}

void Services::init(void)
{
    qwr.init();

    uint32_t              err_code;
    ble_uuid_t            ble_uuid;
    ble_add_char_params_t add_char_params;

    // Add service.
    ble_uuid128_t base_uuid = {S10_UUID_BASE};
    err_code = sd_ble_uuid_vs_add(&base_uuid, &this->uuid_type);
    APP_ERROR_CHECK(err_code);

    ble_uuid.type = this->uuid_type;
    ble_uuid.uuid = S10_UUID_SERVICE;

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &this->service_handle);
    APP_ERROR_CHECK(err_code);

    memset(&add_char_params, 0, sizeof(add_char_params));
    add_char_params.uuid              = S10_UUID_SENSOR_CHAR;
    add_char_params.uuid_type         = this->uuid_type;
    add_char_params.init_len          = 0;
    add_char_params.max_len           = 500; // TODO: change me when updatnig length
    add_char_params.char_props.read   = 1;
    add_char_params.char_props.write  = 1; // write_wo_resp = 1; has been tested, no difference on latency
    add_char_params.char_props.notify = 1;
    add_char_params.char_props.indicate = 1;

    add_char_params.read_access       = SEC_OPEN;
    add_char_params.write_access      = SEC_OPEN;
    add_char_params.cccd_write_access = SEC_OPEN;
    add_char_params.is_var_len        = 1;    

    err_code = characteristic_add(this->service_handle,
                                  &add_char_params,
                                  &this->char_handles);
    APP_ERROR_CHECK(err_code);


}


void Services::send_sensor_data(std::vector<XXXXXXXXX::Measurement> const &sensor_data)
{
    // do not send anything if do not have any data
    if(sensor_data.size() == 0) {
        return;
    }

    SensorData_Msg *msg = new SensorData_Msg();
    
    // setup package
    msg->type = MsgType::SENSOR_DATA;
    // <<<---- NOTE CODE REMOVED HERE TO PREPARE MESSAGE ->>>>
    NRF_LOG_INFO("Send count: %d", sensor_data.size());

    msg->length = sensor_data.size();
    add_message(msg);
}

void s10::Services::add_message(Services::Msg *msg)
{
    msg_queue.push_back(msg);
}


GAP::GAP()
{
}

void GAP::init(void)
{
    ret_code_t err_code;
    ble_gap_conn_params_t gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)DEVICE_NAME,
                                          strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}
/**@brief Function for handling Queued Write Module errors.
 *
 * @details A pointer to this function will be passed to each service which may need to inform the
 *          application about an error.
 *
 * @param[in]   nrf_error   Error code containing information about what went wrong.
 */
void QWR::nrf_qwr_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


/**@brief Function for initializing the GATT module.
 */
void GATT::init(void)
{
    ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_ble_gatt_att_mtu_periph_set(&m_gatt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
Advertising::Advertising(Services *services) : 
            services(services),
            adv_data({
                .adv_data =
                    {
                        .p_data = m_enc_advdata,
                        .len = BLE_GAP_ADV_SET_DATA_SIZE_MAX},
                .scan_rsp_data =
                    {
                        .p_data = m_enc_scan_response_data,
                        .len = BLE_GAP_ADV_SET_DATA_SIZE_MAX

                    }}) 
{


}

void Advertising::init()
{
    ret_code_t err_code;
    ble_advdata_t _advdata;
    ble_advdata_t srdata;

    ble_uuid_t adv_uuids[] = {{S10_UUID_SERVICE, this->services->uuid_type}};

    // Build and set advertising data.
    memset(&_advdata, 0, sizeof(_advdata));

    _advdata.name_type = BLE_ADVDATA_FULL_NAME;
    _advdata.include_appearance = true;
    _advdata.flags = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;

    memset(&srdata, 0, sizeof(srdata));
    srdata.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
    srdata.uuids_complete.p_uuids = adv_uuids;

    err_code = ble_advdata_encode(&_advdata, this->adv_data.adv_data.p_data, &this->adv_data.adv_data.len);
    APP_ERROR_CHECK(err_code);

    err_code = ble_advdata_encode(&srdata, this->adv_data.scan_rsp_data.p_data, &this->adv_data.scan_rsp_data.len);
    APP_ERROR_CHECK(err_code);

    // Initialize advertising parameters (used when starting advertising).
    ble_gap_adv_params_t adv_params;

    // Set advertising parameters.
    memset(&adv_params, 0, sizeof(adv_params));

    adv_params.primary_phy = BLE_GAP_PHY_1MBPS;
    adv_params.duration = APP_ADV_DURATION;
    adv_params.properties.type = BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
    adv_params.p_peer_addr = NULL;
    adv_params.filter_policy = BLE_GAP_ADV_FP_ANY;
    adv_params.interval = APP_ADV_INTERVAL;

    err_code = sd_ble_gap_adv_set_configure(&m_adv_handle, &this->adv_data, &adv_params);
    APP_ERROR_CHECK(err_code);
}




